#include <Kokkos_Core.hpp>
#include <functional>
#include <cassert>
#include <iostream>
#include <utility>
#include <cstring>

#include "Kokkos_Timer.hpp"
#include "icon_loops.hpp"
using icon_loops::outer_loop, icon_loops::inner_loop;

using space_t = Kokkos::DefaultExecutionSpace::memory_space;
typedef Kokkos::MDRangePolicy<Kokkos::DefaultExecutionSpace, Kokkos::IndexType<int>, Kokkos::Rank<3>> md_range_policy;

Kokkos::Timer timer;

// constexpr int nblocks = 2;
// constexpr int nlev = 90;
// constexpr int nproma = 55000;

#define ENABLE_CHECK_BOUNDS
//#define ENABLE_CHECK_BOUNDS_2D

static void validate(double* array, double *ref, int nblocks, int nlev, int nproma) {
  for (int i = 0; i < nblocks * nlev * nproma; ++i) {
    assert(array[i] == ref[i]);
  }
}

KOKKOS_INLINE_FUNCTION void check_bounds(int i1, int i2, int i3, int n1, int n2, int n3) {
#ifdef ENABLE_CHECK_BOUNDS
  assert(i1 >= 0 && i2 >= 0 && i3 >= 0 && i1 < n1 && i2 < n2 && i3 < n3);
#endif
}

KOKKOS_INLINE_FUNCTION void check_bounds_2d(int i1, int i2, int n1, int n2) {
#ifdef ENABLE_CHECK_BOUNDS_2D
  assert(i1 >= 0 && i2 >= 0 && i1 < n1 && i2 < n2 );
#endif
}


void scenario_1(double* array, int nblocks, int nlev, int nproma, bool print = true) {
  if (print)
    std::cout << "scenario 1: Default layout; view(array, nblocks, nlev, nproma); d_view(jb, jk, jc) ----- "
              << std::endl;

  Kokkos::View<double***, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nblocks, nlev, nproma);

  using space_t = Kokkos::DefaultExecutionSpace::memory_space;
  auto d_view   = Kokkos::create_mirror_view_and_copy(space_t(), view);

  timer.reset();

  for (int jb = 0; jb < nblocks; ++jb)
    Kokkos::parallel_for(
        "", Kokkos::RangePolicy<>(0, nproma), KOKKOS_LAMBDA(const int jc) {
          for (int jk = 0; jk < nlev; ++jk) {
            int p = jb * nlev * nproma + jk * nproma + jc;
            check_bounds(jb, jk, jc, d_view.extent(0), d_view.extent(1), d_view.extent(2));
            d_view(jb, jk, jc) = p;
          }
        });
  Kokkos::fence();

  Kokkos::fence();
  if (print) printf("Time = %f ms\n\n", timer.seconds() * 1000);
  Kokkos::deep_copy(view, d_view);

  //validate(array, nblocks, nlev, nproma);
}


void scenario_1b(double* array, int nblocks, int nlev, int nproma) {
  std::cout << "scenario 1b: Default layout right; view(array, nblocks, nlev, nproma); d_view(jb, jk, jc) ----- "
              << std::endl;

  Kokkos::View<double***, Kokkos::LayoutRight, Kokkos::MemoryUnmanaged> d_view(array, nblocks, nlev, nproma);

  timer.reset();

  for (int jb = 0; jb < nblocks; ++jb)
    Kokkos::parallel_for(
        "", Kokkos::RangePolicy<>(0, nproma), KOKKOS_LAMBDA(const int jc) {
          for (int jk = 1; jk < nlev-1; ++jk) {
            check_bounds(jb, jk, jc, d_view.extent(0), d_view.extent(1), d_view.extent(2));
            d_view(jb,jk,jc) = d_view(jb,jk,jc)*0.5 + d_view(jb,jk-1,jc)*0.25 + d_view(jb,jk+1,jc)*0.25;
          }
        });
  Kokkos::fence();

  printf("Time = %f ms\n\n", timer.seconds() * 1000);

}

#if 0
void scenario_1c(double* array, int nblocks, int nlev, int nproma) {
  std::cout << "scenario 1c: Default layout; view(array, ncells, nlev); d_view(jc, jk) ----- "
              << std::endl;
  int ncells = nblocks*nproma;

  Kokkos::View<double**, Kokkos::MemoryUnmanaged> d_view(array, ncells, nlev);

  timer.reset();

    Kokkos::parallel_for(
        "", Kokkos::RangePolicy<>(0, ncells), KOKKOS_LAMBDA(const int jc) {
          for (int jk = 0; jk < nlev; ++jk) {
#if defined(DEMO_DEVICE)
            int p = jk * ncells + jc;
#else
            int p = jc * nlev + jk;
#endif
            check_bounds_2d(jc, jk, d_view.extent(0), d_view.extent(1));
            d_view(jc, jk) = p;
          }
        });
  Kokkos::fence();

  printf("Time = %f ms\n\n", timer.seconds() * 1000);

}

void scenario_2(double* array, int nblocks, int nlev, int nproma, bool print = true) {
  if (print)
    std::cout << "scenario 2: Right layout; view(array, nproma, nlev, nblocks); d_view(jc, jk, jb) ----- " << std::endl;

  Kokkos::View<double***, Kokkos::LayoutRight, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nproma, nlev,
                                                                                                nblocks);

  using space_t = Kokkos::DefaultExecutionSpace::memory_space;
  auto d_view   = Kokkos::create_mirror_view_and_copy(space_t(), view);

  timer.reset();

  for (int jb = 0; jb < nblocks; ++jb)
    Kokkos::parallel_for(
        "", Kokkos::RangePolicy<>(0, nproma), KOKKOS_LAMBDA(const int jc) {
          for (int jk = 0; jk < nlev; ++jk) {
            int p = jc * nlev * nblocks + jk * nblocks + jb;
            check_bounds(jc, jk, jb, d_view.extent(0), d_view.extent(1), d_view.extent(2));
            d_view(jc, jk, jb) = p;

            //  printf("%f ", d_view(jb, jk, jc));
          }
        });

  Kokkos::fence();
  if (print) printf("Time = %f ms\n\n", timer.seconds() * 1000);
  Kokkos::deep_copy(view, d_view);
  //validate(array, nblocks, nlev, nproma);
}

void scenario_2b(double* array, int nblocks, int nlev, int nproma, bool print = true) {
  if (print)
    std::cout << "scenario 2b: Right 2b layout; view(array, nblocks, nlev, nproma); d_view(jb, jk, jc) ----- "
              << std::endl;

  Kokkos::View<double***, Kokkos::LayoutRight, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nblocks, nlev,
                                                                                                nproma);

  using space_t = Kokkos::DefaultExecutionSpace::memory_space;
  auto d_view   = Kokkos::create_mirror_view_and_copy(space_t(), view);

  timer.reset();

  for (int jb = 0; jb < nblocks; ++jb)
    Kokkos::parallel_for(
        "", Kokkos::RangePolicy<>(0, nproma), KOKKOS_LAMBDA(const int jc) {
          for (int jk = 0; jk < nlev; ++jk) {
            int p = jb * nlev * nproma + jk * nproma + jc;
            check_bounds(jb, jk, jc, d_view.extent(0), d_view.extent(1), d_view.extent(2));
            d_view(jb, jk, jc) = p;

            //  printf("%f ", d_view(jb, jk, jc));
          }
        });

  Kokkos::fence();
  if (print) printf("Time = %f ms\n\n", timer.seconds() * 1000);
  Kokkos::deep_copy(view, d_view);
  validate(array, nblocks, nlev, nproma);
}

// slow on CPU
void scenario_3(double* array, int nblocks, int nlev, int nproma, bool print = true) {
  if (print)
    std::cout << "scenario 3: Left layout; view(array, nproma, nlev, nblocks); d_view(jc, jk, jb) ----- " << std::endl;

  Kokkos::View<double***, Kokkos::LayoutLeft, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nproma, nlev,
                                                                                               nblocks);

  using space_t = Kokkos::DefaultExecutionSpace::memory_space;
  auto d_view   = Kokkos::create_mirror_view_and_copy(space_t(), view);

  timer.reset();
  for (int jb = 0; jb < nblocks; ++jb)
    Kokkos::parallel_for(
        "", Kokkos::RangePolicy<>(0, nproma), KOKKOS_LAMBDA(const int jc) {
          for (int jk = 0; jk < nlev; ++jk) {
            int p = jb * nlev * nproma + jk * nproma + jc;
            check_bounds(jc, jk, jb, d_view.extent(0), d_view.extent(1), d_view.extent(2));
            d_view(jc, jk, jb) = p;

            //  printf("%f ", d_view(jb, jk, jc));
          }
        });

  Kokkos::fence();
  if (print) printf("Time = %f ms\n\n", timer.seconds() * 1000);
  Kokkos::deep_copy(view, d_view);
  validate(array, nblocks, nlev, nproma);
}

void scenario_4(double* array, int nblocks, int nlev, int nproma, bool print = true) {
  if (print)
    std::cout << "scenario 4: Default layout; view(array, nproma, nlev, nblocks); d_view(jc, jk, jb) ----- "
              << std::endl;

  Kokkos::View<double***, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nproma, nlev, nblocks);

  using space_t = Kokkos::DefaultExecutionSpace::memory_space;
  auto d_view   = Kokkos::create_mirror_view_and_copy(space_t(), view);

  timer.reset();
  for (int jb = 0; jb < nblocks; ++jb)
    Kokkos::parallel_for(
        "", Kokkos::RangePolicy<>(0, nproma), KOKKOS_LAMBDA(const int jc) {
          for (int jk = 0; jk < nlev; ++jk) {
            int p = jc * nlev * nblocks + jk * nblocks + jb;
            check_bounds(jc, jk, jb, d_view.extent(0), d_view.extent(1), d_view.extent(2));
            d_view(jc, jk, jb) = p;

            //  printf("%f ", d_view(jb, jk, jc));
          }
        });
  Kokkos::fence();
  if (print) printf("Time = %f ms\n\n", timer.seconds() * 1000);
  Kokkos::deep_copy(view, d_view);
  validate(array, nblocks, nlev, nproma);
}

void scenario_4b(double* array, int nblocks, int nlev, int nproma, bool print = true) {
  if (print)
    std::cout << "scenario 4b (Dmitry's solution): view(array, nproma, nlev, nblocks); d_view(jc, jk, jb) ----- "
              << std::endl;

  Kokkos::View<double***, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nproma, nlev, nblocks);

  using space_t   = Kokkos::DefaultExecutionSpace::memory_space;
  auto d_view_tmp = Kokkos::create_mirror_view_and_copy(space_t(), view);
  Kokkos::View<double***, space_t> d_view("d_view", nproma, nlev, nblocks);
  Kokkos::deep_copy(d_view_tmp, view);
  Kokkos::deep_copy(d_view, d_view_tmp);

  timer.reset();
  for (int jb = 0; jb < nblocks; ++jb)
    Kokkos::parallel_for(
        "", Kokkos::RangePolicy<>(0, nproma), KOKKOS_LAMBDA(const int jc) {
          for (int jk = 0; jk < nlev; ++jk) {
            int p              = jc * nlev * nblocks + jk * nblocks + jb;
            d_view(jc, jk, jb) = p;

            //  printf("%f ", d_view(jb, jk, jc));
          }
        });

  Kokkos::fence();
  if (print) printf("Time = %f ms\n\n", timer.seconds() * 1000);
  Kokkos::deep_copy(d_view_tmp, d_view);
  Kokkos::deep_copy(view, d_view_tmp);
  validate(array, nblocks, nlev, nproma);
}

#if defined(DEMO_DEVICE)
  using Layout = Kokkos::LayoutLeft;
#else
  using Layout = Kokkos::LayoutRight;
#endif

template <class T> struct LoopFunctor {
  Kokkos::View<T**, Kokkos::LayoutStride> view;
  int nproma, nlev, nblocks, jb;

  LoopFunctor(int nproma, int nlev, int nblocks, int jb, Kokkos::View<T***, Kokkos::LayoutStride> d_view)
      : nproma(nproma), nlev(nlev), nblocks(nblocks), jb(jb) {
    view = Kokkos::subview(d_view, Kokkos::ALL, Kokkos::ALL, jb);
  }

  KOKKOS_INLINE_FUNCTION
  void operator()(const int jc) const {
    for (int jk = 0; jk < nlev; ++jk) {
#if defined(DEMO_DEVICE)
      int p = jb * nlev * nproma + jk * nproma + jc;
#else
      int p = jc * nlev * nblocks + jk * nblocks + jb;
#endif
      view(jc, jk) = p;
    }
  }
};


void scenario_5(double* array, int nblocks, int nlev, int nproma, bool print = true) {
  if (print) std::cout << "scenario 5: Adaptable layout & functor & subview (array, nproma, nlev);" << std::endl;

  Kokkos::View<double***, Layout, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nproma, nlev, nblocks);

  using space_t = Kokkos::DefaultExecutionSpace::memory_space;
  auto d_view   = Kokkos::create_mirror_view_and_copy(space_t(), view);

  timer.reset();
  for (int jb = 0; jb < nblocks; ++jb) {
    LoopFunctor<double> loop(nproma, nlev, nblocks, jb, d_view);
    Kokkos::parallel_for(Kokkos::RangePolicy<>(0, nproma), loop);
  }

  Kokkos::fence();
  if (print) printf("Time = %f ms\n\n", timer.seconds() * 1000);
  Kokkos::deep_copy(view, d_view);
  validate(array, nblocks, nlev, nproma);
}

int KOKKOS_FUNCTION get_indexes(int nproma) {

  return nproma + 1 ;
}

void scenario_6(double* array, int nblocks, int nlev, int nproma, bool print = true) {
  if (print) std::cout << "scenario 6: Adaptable Layout &  Hierarchical parallelism" << std::endl;

  Kokkos::View<double***, Layout, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nproma, nlev, nblocks);

  using space_t     = Kokkos::DefaultExecutionSpace::memory_space;
  auto d_view       = Kokkos::create_mirror_view_and_copy(space_t(), view);

  using team_policy = Kokkos::TeamPolicy<>;
  using member_type = Kokkos::TeamPolicy<>::member_type;

  timer.reset();

  Kokkos::parallel_for(
      "blocks", team_policy(nblocks, Kokkos::AUTO), KOKKOS_LAMBDA(const member_type& teamMember) {
        const int jb = teamMember.league_rank();

        volatile int nproma_tmp = get_indexes(nproma -  1);  
        Kokkos::parallel_for(Kokkos::TeamThreadRange(teamMember, nproma_tmp), [&](const int jc) {
          // sequential over the levels
          for (int jk = 0; jk < nlev; ++jk) {
#if defined(DEMO_DEVICE)
            int p = jb * nlev * nproma + jk * nproma + jc;
#else 
                            int p = jc * nlev * nblocks + jk * nblocks + jb;
#endif

            d_view(jc, jk, jb) = p;
          }
        });
      });

  Kokkos::fence();

  if (print) printf("Time = %f ms\n\n", timer.seconds() * 1000);
  Kokkos::deep_copy(view, d_view);
  validate(array, nblocks, nlev, nproma);
}

void scenario_7(double* array, int nblocks, int nlev, int nproma, bool print=true) {

    if(print)
        std::cout << "scenario 7: Default layout; view(array, nblocks, nlev, nproma); d_view(jb, jk, jc) ----- " << std::endl;

    Kokkos::View<double***, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nblocks, nlev, nproma);
 
    using space_t = Kokkos::DefaultExecutionSpace::memory_space;
    auto d_view = Kokkos::create_mirror_view_and_copy(space_t(), view);
    
    timer.reset();
    Kokkos::parallel_for("", Kokkos::RangePolicy<>(0, nblocks), KOKKOS_LAMBDA (const int jb) {
      //for (int jb = 0 ; jb < nblocks; ++jb) {
      for (int jc = 0 ; jc < nproma; ++jc) {
        for (int jk = 0; jk < nlev; ++jk) {
            int p = jb * nlev * nproma + jk * nproma + jc;
            check_bounds(jb,jk,jc, d_view.extent(0), d_view.extent(1), d_view.extent(2));
            d_view(jb, jk, jc) = p;
            //  printf("%f ", d_view(jb, jk, jc));      
          }
        }
      });

    Kokkos::fence();

    if(print)
        printf("Time = %f ms\n\n", timer.seconds() * 1000);
    Kokkos::deep_copy(view, d_view);
    validate(array, nblocks, nlev, nproma);
}

void scenario_7b(double* array, int nblocks, int nlev, int nproma) {

    std::cout << "scenario 7b: +ACC Default layout; view(array, nblocks, nlev, nproma); d_view(jb, jk, jc) ----- " << std::endl;

    Kokkos::View<double***, Kokkos::MemoryUnmanaged> d_view(array, nblocks, nlev, nproma);

   timer.reset();
    Kokkos::parallel_for("", Kokkos::RangePolicy<>(0, nblocks), KOKKOS_LAMBDA (const int jb) {
      //for (int jb = 0 ; jb < nblocks; ++jb) {
      for (int jc = 0 ; jc < nproma; ++jc) {
        for (int jk = 0; jk < nlev; ++jk) {
            
#if defined(DEMO_DEVICE)
            int p = jc * nlev * nblocks + jk * nblocks + jb;
#else 
            int p = jb * nlev * nproma + jk * nproma + jc;

#endif
            check_bounds(jb,jk,jc, d_view.extent(0), d_view.extent(1), d_view.extent(2));
            d_view(jb, jk, jc) = p;
            //  printf("%f ", d_view(jb, jk, jc));      
          }
        }
      });

    Kokkos::fence();
    printf("Time = %f ms\n\n", timer.seconds() * 1000);
}

#define outer_lambda [=]
#define inner_lambda KOKKOS_LAMBDA


#ifdef DEMO_DEVICE
#  define outer_for(STR, START, END, LAMBDA) for(int jb=START; jb<END; ++jb) {LAMBDA(jb);};
#  define inner_for(STR, START, END, LAMBDA) Kokkos::parallel_for(STR,  Kokkos::RangePolicy<>(START, END), LAMBDA)
#else
#  define outer_for(STR, START, END, LAMBDA) Kokkos::parallel_for(STR,  Kokkos::RangePolicy<>(START, END), LAMBDA)
#  define inner_for(STR, START, END, LAMBDA) for(int jc=START; jc<END; ++jc) {LAMBDA(jc);};
#endif



void scenario_7bmacro(double* array, int nblocks, int nlev, int nproma) {

    std::cout << "scenario 7bmacro: +ACC CPU:LayoutLeft GPU:LayoutLeft; view(array, nblocks, nlev, nproma); d_view(jb, jk, jc) ----- " << std::endl;

#if defined(DEMO_DEVICE)
    Kokkos::View<double***, Kokkos::LayoutRight, Kokkos::MemoryUnmanaged> d_view(array, nblocks, nlev, nproma);
#else
    Kokkos::View<double***, Kokkos::LayoutRight, Kokkos::MemoryUnmanaged> d_view(array, nblocks, nlev, nproma);
#endif
   timer.reset();
   outer_for("", 0, nblocks, outer_lambda(const int jb){
       inner_for("",0, nproma, inner_lambda(const int jc) {
           for (int jk = 0; jk < nlev; ++jk) {
             int p = jb * nlev * nproma + jk * nproma + jc;
             d_view(jb, jk, jc) = p;
           }
         });
     });
    Kokkos::fence();
    printf("Time = %f ms\n\n", timer.seconds() * 1000);
}

void scenario_7btemplate(double* array, int nblocks, int nlev, int nproma) {

    std::cout << "scenario 7btemplate: +ACC CPU:LayoutLeft GPU:LayoutLeft; view(array, nblocks, nlev, nproma); d_view(jb, jk, jc) ----- " << std::endl;

#if defined(DEMO_DEVICE)
    Kokkos::View<double***, Kokkos::LayoutRight, Kokkos::MemoryUnmanaged> d_view(array, nblocks, nlev, nproma);
#else
    Kokkos::View<double***, Kokkos::LayoutRight, Kokkos::MemoryUnmanaged> d_view(array, nblocks, nlev, nproma);
#endif
   timer.reset();
   outer_loop(nblocks, outer_lambda(const int jb){
       inner_loop(nproma, inner_lambda(const int jc) {
           for (int jk = 0; jk < nlev; ++jk) {
             int p = jb * nlev * nproma + jk * nproma + jc;
             d_view(jb, jk, jc) = p;
           }
         });
     });
    Kokkos::fence();
    printf("Time = %f ms\n\n", timer.seconds() * 1000);
}


void scenario_8(double* array, int nblocks, int nlev, int nproma) {
    std::cout << "scenario 8: +ACC Default Layout as ICON pointer view(array, nproma, nlev, nblocks); d_view(jb, jk, jc) ----- " << std::endl;

    using space_t = Kokkos::DefaultExecutionSpace::memory_space;
    Kokkos::View<double***, Kokkos::MemoryUnmanaged> d_view(array, nproma, nlev, nblocks);


    timer.reset();
    for (int jb = 0; jb < nblocks; ++jb)
    Kokkos::parallel_for(
        "", Kokkos::RangePolicy<>(0, nproma), KOKKOS_LAMBDA(const int jc) {
          for (int jk = 0; jk < nlev; ++jk) {


#if defined(DEMO_DEVICE)
            int p = jb * nlev * nproma + jk * nproma + jc;
#else 
            int p = jc * nlev * nblocks + jk * nblocks + jb;
#endif
            d_view(jc, jk, jb) = p;

            //  printf("%f ", d_view(jb, jk, jc));
          }
        });
    Kokkos::fence();

    printf("Time = %f ms\n\n", timer.seconds() * 1000);

}


void scenario_8macro(double* array, int nblocks, int nlev, int nproma) {
    std::cout << "scenario 8macro: +ACC CPU:LayoutLeft GPU:LayoutLeft as ICON pointer view(array, nproma, nlev, nblocks); d_view(jc, jk, jb) ----- " << std::endl;
    using space_t = Kokkos::DefaultExecutionSpace::memory_space;
#if defined(DEMO_DEVICE)
    Kokkos::View<double***, Kokkos::LayoutLeft> d_view(array, nproma, nlev, nblocks);
#else
    Kokkos::View<double***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged> d_view(array, nproma, nlev, nblocks);
#endif
    timer.reset();
    outer_for("", 0, nblocks, outer_lambda(const int jb) {
        inner_for("",0, nproma, inner_lambda(const int jc) {
          for (int jk = 0; jk < nlev; ++jk) {
            int p = jb * nlev * nproma + jk * nproma + jc;
            d_view(jc, jk, jb) = p;
            //  printf("%f ", d_view(jb, jk, jc));
          }
        });
      });
    Kokkos::fence();

    printf("Time = %f ms\n\n", timer.seconds() * 1000);

}

void scenario_8template(double* array, int nblocks, int nlev, int nproma) {
    std::cout << "scenario 8template: +ACC CPU:LayoutLeft GPU:LayoutLeft as ICON pointer view(array, nproma, nlev, nblocks); d_view(jc, jk, jb) ----- " << std::endl;
    using space_t = Kokkos::DefaultExecutionSpace::memory_space;
#if defined(DEMO_DEVICE)
    Kokkos::View<double***, Kokkos::LayoutLeft> d_view(array, nproma, nlev, nblocks);
#else
    Kokkos::View<double***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged> d_view(array, nproma, nlev, nblocks);
#endif
    timer.reset();
    outer_loop(nblocks, outer_lambda(const int jb) {
        inner_loop(nproma, inner_lambda(const int jc) {
          for (int jk = 0; jk < nlev; ++jk) {
            int p = jb * nlev * nproma + jk * nproma + jc;
            d_view(jc, jk, jb) = p;
            //  printf("%f ", d_view(jb, jk, jc));
          }
        });
      });
    Kokkos::fence();

    printf("Time = %f ms\n\n", timer.seconds() * 1000);

}

#endif


void openacc_calls(double* array, double *ref, int nblocks, int nlev, int nproma, 
                      std::function<void(double*, int, int, int)> func){
  #if defined(DEMO_DEVICE)
    #pragma acc enter data copyin(array[0:nblocks*nlev*nproma]) 
    {
      #pragma acc host_data use_device(array) 
      {
        func(array, nblocks, nlev, nproma);
      }
    }
    #pragma acc exit data copyout(array[0:nblocks*nlev*nproma])
  #else
      func(array, nblocks, nlev, nproma);
  #endif
      validate(array, ref, nblocks, nlev, nproma);
   
}

void init_array(double *array, int nproma, int nlev, int nblocks) {
  Kokkos::View<double***, Kokkos::LayoutLeft, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nproma, nlev, nblocks);

  for (int jb=0; jb < nblocks; ++jb) {
    for (int jk=0; jk < nlev; ++jk) {
      for (int jc=0; jc < nproma; ++jc) {
        view(jc,jk,jb) = jc + jk*nproma + jb * nproma*nblocks;
      }
    }
  }
}

void physop(double *array, int nproma, int nlev, int nblocks) {
  Kokkos::View<double***, Kokkos::LayoutLeft, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nproma, nlev, nblocks);

  for (int jb=0; jb < nblocks; ++jb) {
    for (int jc=0; jc < nproma; ++jc) {
      for (int jk=1; jk < nlev-1; ++jk) {
        view(jc,jk,jb) = view(jc,jk,jb)*0.5 + view(jc,jk-1,jb)*0.25 + view(jc,jk+1,jb)*0.25;
      }
    }
  }

}


int main() {
  int ncells  = atoi(std::getenv("NCELLS"));
  int nlev    = atoi(std::getenv("NLEV"));
  int nproma  = atoi(std::getenv("NPROMA"));
  int nblocks = (ncells - 1) / nproma + 1;

  std::cout << "nblocks=" << nblocks << ", nlev=" << nlev << ", nproma=" << nproma << std::endl;

  double array_out_ref[nblocks * nlev * nproma];
  double array_in_ref[nblocks * nlev * nproma];
  double array[nblocks * nlev * nproma];
  const size_t nbytes = nproma*nlev*nblocks*sizeof(double);
  Kokkos::initialize();
  {

    init_array(array_in_ref, nproma,nlev,nblocks);

    memcpy(array_out_ref, array_in_ref, nbytes);

    physop(array_out_ref, nproma,nlev,nblocks);

    memcpy(array, array_in_ref, nbytes);

    scenario_1(array, nblocks, nlev, nproma, false);

    //memcpy(array, array_in_ref, nbytes);
    //scenario_1(array, nblocks, nlev, nproma);
    //validate(array, array_out_ref, nblocks, nlev, nproma);

    memcpy(array, array_in_ref, nbytes);
    std::function<void(double*, int, int, int)> s_1b = scenario_1b;
    openacc_calls(array, array_out_ref, nblocks, nlev, nproma, s_1b);
#if 0    


    memset(array, 0.0, sizeof(array));
    std::function<void(double*, int, int, int)> s_1c = scenario_1c;
    openacc_calls(array, nblocks, nlev, nproma, s_1c);

    memset(array, 0.0, sizeof(array));
    std::function<void(double*, int, int, int)> s_7b = scenario_7b;
    openacc_calls(array, nblocks, nlev, nproma, s_7b);

    memset(array, 0.0, sizeof(array));
    std::function<void(double*, int, int, int)> s_7bmacro = scenario_7bmacro;
    openacc_calls(array, nblocks, nlev, nproma, s_7bmacro);

    memset(array, 0.0, sizeof(array));
    std::function<void(double*, int, int, int)> s_7btemplate = scenario_7btemplate;
    openacc_calls(array, nblocks, nlev, nproma, s_7btemplate);

    memset(array, 0.0, sizeof(array));
    std::function<void(double*, int, int, int)> s_8 = scenario_8;
    openacc_calls(array, nblocks, nlev, nproma, s_8);

    memset(array, 0.0, sizeof(array));
    std::function<void(double*, int, int, int)> s_8macro = scenario_8macro;
    openacc_calls(array, nblocks, nlev, nproma, s_8macro);

    memset(array, 0.0, sizeof(array));
    std::function<void(double*, int, int, int)> s_8template = scenario_8template;
    openacc_calls(array, nblocks, nlev, nproma, s_8template);
#endif
/*
    scenario_2(array, nblocks, nlev, nproma);
    scenario_2b(array, nblocks, nlev, nproma);
    scenario_3(array, nblocks, nlev, nproma);
    scenario_4(array, nblocks, nlev, nproma);
    scenario_4b(array, nblocks, nlev, nproma);
    scenario_5(array, nblocks, nlev, nproma);
    scenario_6(array, nblocks, nlev, nproma);
    scenario_7(array, nblocks, nlev, nproma);
*/

  }
  Kokkos::finalize();

  return 0;
}
